/*********************************************************************
* Copyright (c) Impulse Accelerated Technologies, Inc.
* All Rights Reserved.
*/

#include "co.h"
#include "co_error.h"
#include "ants_trigger.h"
#include "co_bits.h"

// Software process declarations (see *_sw.c)
extern void ProducerMac(co_stream from_mac, co_signal start);
extern void ConsumerMac(co_stream to_mac);
IF_SIM(extern void host_proc( co_stream config_strm, co_stream payload_strm, co_signal start_mac);)
IF_NSIM(extern void host_proc( co_stream config_strm, co_stream trigger_strm);)
IF_SIM(extern void print_av_st(const char* str, uint8 sop, uint8 eop, uint8 empty, co_uint256* data);)

#define NUM_MATCHES_NEEDED	7

// ANTS demo defaults
#define MATCH_NUM_OF_BYTE     7  // must be <=8
#define MATCH_STR             0x686974206D6520ULL  // "hit me "  (64-bit maximum)

#define MSG_BYTE_PER_CYCLE    32
#define MATCH_START_OFFSET    0
#define MSG_ID_OFFSET         9

#define MATCH_VAL_POS_HIBIT   ((MSG_BYTE_PER_CYCLE - MATCH_START_OFFSET)*8 - 1)
#define MATCH_VAL_POS_LOWBIT  ((MSG_BYTE_PER_CYCLE - MATCH_START_OFFSET)*8 - MATCH_NUM_OF_BYTE*8)
#define CONN_ID_POS_HIBIT     ((MSG_BYTE_PER_CYCLE - MATCH_NUM_OF_BYTE)*8 - 1)
#define CONN_ID_POS_LOWBIT    (CONN_ID_POS_HIBIT-7)
#define MSG_ID_POS_HIBIT      ((MSG_BYTE_PER_CYCLE - MSG_ID_OFFSET)*8 - 1)
#define MSG_ID_POS_LOWBIT     (MSG_ID_POS_HIBIT-7)


// globals
//static co_uint48 MatchMac = 0;
static co_uint8 MatchIpv4Proto = MCAST_PROTO;
static co_uint32 MatchIPv4Dest = MCAST_IP;
static co_uint16 MatchUdpPort = MCAST_PORT;

// Stream order:
//    0: <command[31:0]>
// 1..n: <data[31:0]>
void ants_config_proc(co_stream cmd_strm, co_stream stat_strm)
{
	IF_SIM(co_error err = co_err_none;)
   co_uint32 dataIn, dataOut;
	co_uint8 cmd;

	IF_SIM(printf("ants_config_proc:Starting\n");)

	co_stream_open(cmd_strm, O_RDONLY, UINT_TYPE(STREAM_COMMAND_WIDTH));
	co_stream_open(stat_strm, O_WRONLY, UINT_TYPE(STREAM_STATUS_WIDTH));
	while (1) {
      IF_SIM(err = )
	   co_stream_read(cmd_strm, &dataIn, sizeof(dataIn));
		IF_SIM(if (err != co_err_none) break;)  // EOS received, end SW sim

      // check for and sync to command pattern
      cmd = dataIn;
      if ((dataIn & 0xffffff00) == STREAM_CMD_MAGIC_NUMBER)
      {
         co_stream_read(cmd_strm, &dataIn, sizeof(dataIn));
         switch(cmd)
         {
            case STREAM_CMD_PING:
               // <passback[31:0]>
               dataOut = dataIn;
               co_stream_write(stat_strm, &dataOut, sizeof(dataOut));
               break;
            case STREAM_CMD_WREG_IPV4DEST:
               // <DestIPv4[31:0]>
               MatchIPv4Dest = dataIn;
               break;
            case STREAM_CMD_WREG_IPV4PROTO:
               // <unused[31:8]><DestIPv4Proto[7:0]>
               MatchIpv4Proto = (co_uint8)dataIn;
               break;
            case STREAM_CMD_WREG_UDPPORT:
               // <unused[31:16]><DestUDPPort[15:0]>
               MatchUdpPort = (co_uint16)dataIn;
               break;
         }
      }
	}

	co_stream_close(cmd_strm);
	co_stream_close(stat_strm);

	IF_SIM(printf("ants_config_proc:Exiting\n");)
}


/*
 * This is the trigger process that looks for a specific data pattern in the payload
 *  data and then sends out a trigger word on match
 */
void ants_trigger_proc(co_stream payload_strm, co_stream trigger_strm)
{
   IF_SIM(co_error err = co_err_none;)
   STREAM_PAYLOAD_TYPE payloadWord;
   STREAM_PAYLOAD_DATA_TYPE payloadData;
   co_uint8 inWordCount;
   AV_ST_EMPTY_TYPE empty;

   co_uint8 msg_id = 0;
   co_uint8 conn_id = 0;
   STREAM_TRIGGER_MSGINFO_TYPE msg_info = 0;

   IF_SIM(printf("ants_trigger_proc:Starting\n");)

   co_stream_open(payload_strm, O_RDONLY, UINT_TYPE(STREAM_PAYLOAD_WIDTH));
   co_stream_open(trigger_strm, O_WRONLY, UINT_TYPE(STREAM_TRIGGER_WIDTH));

   while (1)
   {
      while(1)
      {
         #pragma CO PIPELINE
         co_uint56 dataStr;  // type must match length of data string of interest

         if (co_stream_read_nb(payload_strm, &payloadWord, sizeof(payloadWord)))
         {
            STREAM_TRIGGER_TYPE trigWord;

            STREAM_PAYLOAD_GET_FIELD(empty, payloadWord, EMPTY);
            STREAM_PAYLOAD_GET_FIELD(inWordCount, payloadWord, WORDNUM);
            STREAM_PAYLOAD_GET_FIELD(payloadData, payloadWord, DATA);

            dataStr = 0;
            // looking for string, first bytes are network order and at top of data
            CO_BITS_GET_U(dataStr, payloadData, (STREAM_PAYLOAD_DATA_WIDTH-(MATCH_NUM_OF_BYTE*8)), MATCH_NUM_OF_BYTE*8);

            if ((MATCH_STR == dataStr) & (0 == inWordCount))
            {
               /* output trigger word:
                *    output logic [7:0] msg_id_out,
                *    output logic [7:0] conn_id_out,
                *    output logic [31:0] msg_info_out,
                */
               CO_BITS_GET_U(conn_id, payloadWord, CONN_ID_POS_LOWBIT, (CONN_ID_POS_HIBIT - CONN_ID_POS_LOWBIT + 1));
               conn_id -= 0x30;
               CO_BITS_GET_U(msg_id, payloadWord, MSG_ID_POS_LOWBIT, (MSG_ID_POS_HIBIT - MSG_ID_POS_LOWBIT + 1));
               msg_id -= 0x30;

               // pack trigger data: <msg_id[7:0]><connid[7:0]><msg_info[31:0]>
               STREAM_TRIGGER_SET_FIELD(trigWord, msg_id, MSGID);
               STREAM_TRIGGER_SET_FIELD(trigWord, conn_id, CONNID);
               STREAM_TRIGGER_SET_FIELD(trigWord, msg_info, MSGINFO);

               msg_info++;
               co_stream_write(trigger_strm, &trigWord, sizeof(trigWord));
            }
         }
         IF_SIM( // slow down for software simulation
         else
         {
            Sleep(1);
         }
         )

         IF_SIM(err = co_stream_eos(payload_strm) ? co_err_eos : co_err_none; )
         IF_SIM(if (err != co_err_none) break;)  // EOS received, end SW sim
      }
      IF_SIM(if (err != co_err_none) break;)  // EOS received, end SW sim
   }

   co_stream_close(payload_strm);
   co_stream_close(trigger_strm);

   IF_SIM(printf("ants_trigger_proc:Exiting\n");)
}

/*
 * Packet filter process that filters incoming packets from MAC looking for specific packet
 *  type and destination.  On match, the payload of the matching packet is sent along.
 *  Filtering is in-line with MAC stream so unmodified packets are always sent out.
 */
void ants_filter_proc(co_stream from_mac_strm, co_stream to_mac_strm, co_stream payload_strm)
{
	STREAM_RAW_MAC_TYPE fromMacWord;
	STREAM_RAW_MAC_TYPE toMacWord;
	AV_ST_DATA_TYPE data;
	AV_ST_EOP_TYPE eop;
	AV_ST_SOP_TYPE sop;
	AV_ST_EMPTY_TYPE empty;
//	AV_ST_ERROR_TYPE error;
	STREAM_PAYLOAD_DATA_TYPE payloadWord, lastPayloadWord, veryLastPayloadWord, tmpPayloadHigh, tmpPayloadLow;
	STREAM_PAYLOAD_TYPE payloadWordOut;
	STREAM_PAYLOAD_INFO_TYPE payloadInfo, veryLastPayloadInfo;
	co_uint8 inWordCount, payloadWordCount;
	IF_SIM(co_error err = co_err_none;)
	co_uint4 matchCount = 0;
	co_uint1 matchET,matchHL,matchFO,match4P,matchIPl,matchIPu,matchUP;
	co_uint2 sendLastPayload,sendPayload;

	co_uint32 pktCount = 0;

	IF_SIM(printf("ants_filter_proc:Starting\n");)

	co_stream_open(from_mac_strm, O_RDONLY, UINT_TYPE(STREAM_RAW_MAC_WIDTH));
	co_stream_open(to_mac_strm, O_WRONLY, UINT_TYPE(STREAM_RAW_MAC_WIDTH));
	co_stream_open(payload_strm, O_WRONLY, UINT_TYPE(STREAM_PAYLOAD_WIDTH));

	ZERO_VERYLONG_WORD(fromMacWord);
	inWordCount = 0;
	payloadWordCount = 0;
	sop = 0;
	eop = 0;
	empty = 0;
//	error = 0;
	sendPayload = 0;
	sendLastPayload = 0;
	while (1)
	{
		inWordCount = 0;
		sop = 0;
		eop = 0;
		empty = 0;
//		error = 0;

		while(1)
		{
			#pragma CO PIPELINE
			if (co_stream_read_nb(from_mac_strm, &fromMacWord, sizeof(fromMacWord)))
			{
            co_uint16 ethType;
            co_uint4 hdrLen;
            co_uint16 fragOffset;
            co_uint8 dataIpv4Proto;
            co_uint32 dataIPv4Dest;
            co_uint16 dataUdpPort;
            co_uint16 tmp16;
            co_uint16 udpLen;

            // read fields from incoming word
            AV_ST_GET_FIELD(data, fromMacWord, DATA);
            AV_ST_GET_FIELD(eop, fromMacWord, EOP);
            AV_ST_GET_FIELD(sop, fromMacWord, SOP);
            AV_ST_GET_FIELD(empty, fromMacWord, EMPTY);
   //			AV_ST_GET_FIELD(fromMacWord, error, ERROR);
            IF_SIM(

            print_av_st("ants_filter_proc:Read from src ",sop,eop,empty,&data);
            )

            // sop is first word in packet
            if (sop)
            {
               inWordCount = 0;
               matchCount = 0;
               payloadWordCount = 0xff;
               ZERO_VERYLONG_WORD(payloadWord);
               ZERO_VERYLONG_WORD(lastPayloadWord);
            }

            CO_BITS_GET_U(ethType, data, (256-(14*8)), 16); // 2 bytes word #0 are Ethernet type (IPV4:0x0800)
            CO_BITS_GET_U(hdrLen, data, (256-(15*8)), 4);   // nibble word #0 is header length
            CO_BITS_GET_U(fragOffset, data, (256-(22*8)), 14); // 14 bits #0 are fragment offset and more fragments bit
            CO_BITS_GET_U(dataIpv4Proto, data, (256-(24*8)), 8); // byte word #0 is IPV4 protocol
            CO_BITS_GET_U(tmp16, data, (256-(32*8)), 16); // 2 bytes word #0 are upper half of IPv4 dest address

            dataIPv4Dest = (co_uint32)tmp16 << 16;

            //Check not supported features // {no_vlan_tag , fixed IP header length ,no_fragment}
            matchET = (0 == inWordCount) & (0x0800 == ethType);
            matchHL = (0 == inWordCount) & (5 == hdrLen);
            matchFO = (0 == inWordCount) & (0 == fragOffset);
            match4P = (0 == inWordCount) & (MatchIpv4Proto == dataIpv4Proto);
            matchIPu = (0 == inWordCount) & (tmp16 == (co_uint16)(MatchIPv4Dest >> 16));

            CO_BITS_GET_U(dataUdpPort, data, (256-(6*8)), 16); // 2 bytes word #1 are UDP dest port
            matchUP = (1 == inWordCount) & (MatchUdpPort == dataUdpPort);
            CO_BITS_GET_U(tmp16, data, (256-(2*8)), 16); // 2 bytes word #1 are lower half of IPv4 dest address
            matchIPl = (1 == inWordCount) & (tmp16 == (co_uint16)MatchIPv4Dest);

            if (1 == inWordCount)
            {
               CO_BITS_GET_U(udpLen, data, (256-(8*8)), 16); // 2 bytes word #1 are UDP length
            }

            // final sum of individual checks creates faster logic tree vs "+="
            matchCount += (co_uint4)matchET + (co_uint4)matchHL + (co_uint4)matchFO + (co_uint4)match4P
                           + (co_uint4)matchIPl + (co_uint4)matchIPu + (co_uint4)matchUP;

            // piece together payload into re-aligned words
            if ((inWordCount <= 1) & (eop != 0))
            {
               CO_BITS_GET_U(tmpPayloadHigh, data, 0, PAYLOAD_BYTES_FIRST_WORD*8);         // upper payload starts in lower PAYLOAD_BYTES_FIRST_WORD bytes
               ZERO_VERYLONG_WORD(tmpPayloadLow);  // ensure 0's in unused bytes
            }
            else
            {
               CO_BITS_GET_U(tmpPayloadHigh, lastPayloadWord, 0, PAYLOAD_BYTES_FIRST_WORD*8);         // upper payload starts in lower PAYLOAD_BYTES_FIRST_WORD bytes
               CO_BITS_GET_U(tmpPayloadLow, data, 256-(PAYLOAD_BYTES_REST_WORDS*8), PAYLOAD_BYTES_REST_WORDS*8);   // lower payload ends in next upper PAYLOAD_BYTES_REST_WORDS bytes
            }

            // assemble full payload word
            CO_BITS_PUT_U(payloadWord, tmpPayloadLow, 0, PAYLOAD_BYTES_REST_WORDS*8);  // use last lower bytes
            CO_BITS_PUT_U(payloadWord, tmpPayloadHigh, PAYLOAD_BYTES_REST_WORDS*8, PAYLOAD_BYTES_FIRST_WORD*8); // use current upper bytes

            if ((matchCount >= NUM_MATCHES_NEEDED) & ((inWordCount > 1) | (eop != 0)))
            {
               sendPayload = 1;
               payloadWordCount++;
            }

            // handle last packet word
            COPY_VERYLONG_WORD(veryLastPayloadWord, tmpPayloadHigh);
            if ((matchCount >= NUM_MATCHES_NEEDED) & (eop))
            {
               if (empty >= PAYLOAD_BYTES_FIRST_WORD)
               {
                  payloadInfo = ((STREAM_PAYLOAD_INFO_TYPE)(empty - PAYLOAD_BYTES_FIRST_WORD) << 8) | payloadWordCount;
                  veryLastPayloadInfo = ((STREAM_PAYLOAD_INFO_TYPE)(empty - PAYLOAD_BYTES_FIRST_WORD) << 8) | (STREAM_PAYLOAD_INFO_TYPE)((co_uint8)(payloadWordCount + 1));
               }
               else
               {
                  payloadInfo = ((STREAM_PAYLOAD_INFO_TYPE)(empty + PAYLOAD_BYTES_REST_WORDS) << 8) | payloadWordCount;
                  veryLastPayloadInfo = ((STREAM_PAYLOAD_INFO_TYPE)(empty + PAYLOAD_BYTES_REST_WORDS) << 8) | (STREAM_PAYLOAD_INFO_TYPE)((co_uint8)(payloadWordCount + 1));
               }
               // watch for one more word if it needs to go out
               if ((empty > 0) & (empty < PAYLOAD_BYTES_FIRST_WORD) & (inWordCount > 1))
               {
                  sendPayload = 2;
                  payloadInfo &= 0x00ff;  // clear empty bits for full word just before last
               }
            }
            else
            {
               payloadInfo = payloadWordCount;  // empty = 0
            }
            COPY_VERYLONG_WORD(lastPayloadWord, data);

            COPY_VERYLONG_WORD(toMacWord,fromMacWord);

            co_stream_write(to_mac_strm, &toMacWord, sizeof(toMacWord));
            inWordCount++;
			}
	      IF_SIM( else // slow down for software simulation
	      {
	         Sleep(1);
	      })

	      if (sendPayload)
	      {
	         if (sendLastPayload)
	         {
	            COPY_VERYLONG_WORD(payloadWord, veryLastPayloadWord);
	            payloadInfo = veryLastPayloadInfo;
	         }

	         if (2 == sendPayload)
	         {
	            sendLastPayload = 1;
	         }
	         else
	         {
	            sendLastPayload = 0;
	         }

            pktCount++;
            STREAM_PAYLOAD_SET_FIELD(payloadWordOut, payloadWord, DATA);
            STREAM_PAYLOAD_SET_FIELD(payloadWordOut, payloadInfo, INFO);

            co_stream_write(payload_strm, &payloadWordOut, sizeof(payloadWordOut));
	         sendPayload--;
	      }
			IF_SIM(err = co_stream_eos(from_mac_strm) ? co_err_eos : co_err_none; )
			IF_SIM(if ((sendPayload == 0) && (err != co_err_none)) break;)  // EOS received, end SW sim
		}
		IF_SIM(if (err != co_err_none) break;)  // EOS received, end SW sim
	}
	co_stream_close(to_mac_strm);
	co_stream_close(from_mac_strm);
	co_stream_close(payload_strm);

	IF_SIM(printf("ants_filter_proc:Exiting\n");)
}

//
// Impulse C configuration function
//
void config_ants_trigger(void *arg)
{
	co_stream mac_in_strm;
	co_stream mac_out_strm;
	co_stream command_strm;
	co_stream status_strm;
	co_stream trigger_strm;
	co_stream payload_strm;

	co_process ants_filter_process;
	co_process ants_trigger_process;
	co_process ants_config_process;
	co_process host_process;

	IF_SIM(
	co_process consumer_mac_process;
	co_process producer_mac_process;
	co_signal start_sig;
	)
	IF_NSIM(
	co_port mac_in_strm_port;
	co_port mac_out_strm_port;
	co_port trigger_strm_port;
	co_port command_strm_port;
	co_port status_strm_port;
	)

	mac_in_strm = co_stream_create("mac_in_strm", UINT_TYPE(STREAM_RAW_MAC_WIDTH), STREAM_RAW_MAC_DEPTH);
	mac_out_strm = co_stream_create("mac_out_strm", UINT_TYPE(STREAM_RAW_MAC_WIDTH), STREAM_RAW_MAC_DEPTH);
	command_strm = co_stream_create("command_strm", UINT_TYPE(STREAM_COMMAND_WIDTH), STREAM_COMMAND_DEPTH);
	status_strm = co_stream_create("status_strm", UINT_TYPE(STREAM_STATUS_WIDTH), STREAM_STATUS_DEPTH);
	trigger_strm = co_stream_create("trigger_strm", UINT_TYPE(STREAM_TRIGGER_WIDTH), STREAM_TRIGGER_DEPTH);
	payload_strm = co_stream_create("payload_strm", UINT_TYPE(STREAM_PAYLOAD_WIDTH), STREAM_PAYLOAD_DEPTH);

	IF_NSIM(
	// create ports for Avalon-ST connections in Qsys
	mac_out_strm_port = co_port_create("mac_out_strm_port", co_output, mac_out_strm);
	mac_in_strm_port = co_port_create("mac_in_strm_port", co_input, mac_in_strm);

	trigger_strm_port = co_port_create("trigger_strm_port", co_output, trigger_strm);
	command_strm_port = co_port_create("command_strm_port", co_input, command_strm);
	status_strm_port = co_port_create("status_strm_port", co_output, status_strm);

	// Avalon-ST without EOS support and accepting Avalon-ST control signals (SOP/EOP, Empty, and Error bits)
	//  as part of co_stream's data
	co_port_config(mac_out_strm_port, co_loc, "avalon_st_raw_port");
	co_port_config(mac_in_strm_port, co_loc, "avalon_st_raw_port");

	co_port_config(trigger_strm_port, co_loc, "avalon_st_raw_port");
	co_port_config(command_strm_port, co_loc, "avalon_st_raw_port");
	co_port_config(status_strm_port, co_loc, "avalon_st_raw_port");
	)

	// All MAC ports don't have software processes
	IF_SIM(
	start_sig = co_signal_create("start_sig");
	producer_mac_process = co_process_create(
		"ProducerMac",
		(co_function)ProducerMac,
		2,
		mac_in_strm,
		start_sig
	);
	consumer_mac_process = co_process_create(
		"ConsumerMac",
		(co_function)ConsumerMac,
		1,
		mac_out_strm
	);
	)

	// host software processes
	IF_SIM(
	host_process = co_process_create(
		"host_proc",
		(co_function)host_proc,
		4,
		command_strm,
		status_strm,
		trigger_strm,
		start_sig
	);
	)
//	IF_NSIM(
//	host_process = co_process_create(
//		"host_proc",
//		(co_function)host_proc,
//		2,
//      config_strm,
//      trigger_strm
//	);
//	)
	// hardware process
	ants_config_process = co_process_create(
	      "ants_config_proc",
	      (co_function)ants_config_proc,
	      2,
	      command_strm,
	      status_strm
	   );

	ants_filter_process = co_process_create(
		"ants_filter_proc",
		(co_function)ants_filter_proc,
		3,
		mac_in_strm,
		mac_out_strm,
		payload_strm
	);

   ants_trigger_process = co_process_create(
      "ants_trigger_proc",
      (co_function)ants_trigger_proc,
      2,
      payload_strm,
      trigger_strm
   );

   // Use "wire" (depth=0) type co_streams to minimize latency to 1 cycle, FIFOs are 2-3 cycles
   // NOTE: Requires Altera Qsys PSP patch file
   co_stream_config(mac_in_strm, co_kind, "wire");
   co_stream_config(mac_out_strm, co_kind, "wire");
   co_stream_config(command_strm, co_kind, "wire");
   co_stream_config(payload_strm, co_kind, "wire");
   co_stream_config(trigger_strm, co_kind, "wire");

	co_process_config(ants_filter_process, co_loc, "pe0");
	co_process_config(ants_trigger_process, co_loc, "pe0");
	co_process_config(ants_config_process, co_loc, "pe0");
}

co_architecture co_initialize(int param)
{
	return(co_architecture_create("ants_trigger_arch", "Solarflare", config_ants_trigger, (void *)param));
}

