/**************************************************************************\
 *    Filename: ants_hal_api.c
 * Description: ANTS hal api file
 *
 * Copyright (c) 2015-2016, Solarflare Communications Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * \**************************************************************************/

#include "aoeants.h"
#include "ants_hal_api.h"

#define SF_HW_BLOCK_REG_WIDTH (sizeof(uint32_t))

/* Memory size */
#define MSG_BODY_MEM_SIZE  0x800
#define MSG_LEN_CSUM_MEM_SIZE 0x4
#define PER_CONNECTION_ANTS_ETH_HDR_MEM_SIZE 0x20
#define PER_CONNECTION_ANTS_TCP_HDR_MEM_SIZE 0x40
#define PER_CONNECTION_ANTS_IPV4_HDR_MEM_SIZE 0x40
#define PER_CONNECTION_ANTS_TCP_RX_TABLE_SIZE 0x8
#define PER_CONNECTION_SEQ_AND_SEQ_NUM_MAX_SIZE 0xF

/* Local Macro Definations */
#define WORD_SIZE sizeof(uint32_t)
#define MSG_LEN_MASK 0x10
#define TCP_ACK_OFFSET 0x8
#define SEQ_NUM_OFFSET 0x4
#define ETH_DA_SA_SIZE 0xC
#define TCP_CSUM_OFFSET 0x2
#define IPV4_CSUM_OFFSET 0xa
#define TCP_WINDOW_OFFSET 0x4
#define ETHER_TYPE_OFFSET 0x10
#define LOW_TWO_BYTE_MASK 0x0000FFFF
#define HIGH_TWO_BYTE_MASK 0xFFFF0000
#define FIXED_PREFIX_LENGTH 0x2a
#define MSG_PAYLOAD_MAX_SIZE 0x800
#define TCP_WINDOW_SIZE_OFFSET 0xe
#define LEFT_SHIFT_LOW_TWO_BYTE 0x10
#define TCP_ENABLE_DISABLE_OFFSET 0x8

/*
 * @brief Per connection TCP entry update of seq_num_max ACK , window size and tcp checksum
 *
 * @param handle <amm_handle_t> Handle to the device
 * @param resource_handle <uint32_t> Connection Id
 * @param tcp_conn <void*> Pointer to per TCP connection entry
 *
 * @return AMM_SUCCESS or different error codes
 */
amm_err_t antshal_tcp_connection_entry_update(amm_handle_t handle, uint32_t resource_handle, void *tcp_conn)
{
  uint32_t data = 0;
  uint32_t seq_num_mem_addr = 0;
  uint32_t ants_tcp_rx_table_mem_addr = 0;
  ants_tcp_connection_entry_info_t *tcp_connection = NULL;
  amm_err_t status = AMM_SUCCESS;

/* Parameter Validation */
  if(NULL == handle)
    return AMM_ERROR_NOT_OPEN;

  if((NULL == tcp_conn) || ((int32_t)resource_handle < 0))
    return AMM_ERROR_INVALID_PARAM;

  tcp_connection = (ants_tcp_connection_entry_info_t*)tcp_conn;
  seq_num_mem_addr = SF_SEQ_MAX_SEQ_NUM_REG + ( resource_handle * PER_CONNECTION_SEQ_AND_SEQ_NUM_MAX_SIZE );
  ants_tcp_rx_table_mem_addr = SF_ANTS_TCP_RX_TABLE_REG + ( resource_handle * PER_CONNECTION_ANTS_TCP_RX_TABLE_SIZE );

  /*Seq_num_max configuration*/
  status = ants_write_reg(handle, seq_num_mem_addr, tcp_connection->seq_num_max);
  if(status != AMM_SUCCESS) goto error;

  /*ACK, window and tcp checksum configuration*/
  data=0;
  data = *((uint16_t*)(tcp_connection->header + tcp_connection->tcp_header_offset + TCP_WINDOW_SIZE_OFFSET));
  data |= (tcp_connection->tcp_header_checksum << LEFT_SHIFT_LOW_TWO_BYTE);

  status = ants_write_reg(handle, ants_tcp_rx_table_mem_addr , *((uint32_t*)(tcp_connection->header + tcp_connection->tcp_header_offset + TCP_ACK_OFFSET))) ;
  if(status != AMM_SUCCESS) goto error;

  status = ants_write_reg(handle, ants_tcp_rx_table_mem_addr + TCP_WINDOW_OFFSET , data);
  if(status != AMM_SUCCESS) goto error;

error:
  return status;
}

/*
 * @brief Per connection TCP entry programming of ETH,IPv4 and TCP header
 *
 * @param handle <amm_handle_t> Handle to the device
 * @param resource_handle <uint32_t> Connection Id
 * @param tcp_conn <void*> Pointer to per TCP connection entry
 *
 * @return AMM_SUCCESS or different error codes
 */
amm_err_t antshal_tcp_connection_entry(amm_handle_t handle, uint32_t resource_handle, void *tcp_conn)
{
  uint32_t eth_hdr_mem_addr = 0;
  uint32_t ipv4_hdr_mem_addr = 0;
  uint32_t tcp_hdr_mem_addr = 0;
  uint32_t seq_num_mem_addr = 0;
  uint32_t ants_tcp_rx_table_mem_addr = 0;
  uint32_t data = 0 , seq = 0, ack = 0;
  ants_tcp_connection_entry_info_t *tcp_connection = NULL;

  amm_err_t status = AMM_SUCCESS;

  /* Parameter Validation */
  if(NULL == handle)
    return AMM_ERROR_NOT_OPEN;

  if((NULL == tcp_conn) || ((int32_t)resource_handle < 0))
    return AMM_ERROR_INVALID_PARAM;

  tcp_connection = (ants_tcp_connection_entry_info_t*)tcp_conn;
  eth_hdr_mem_addr = SF_ANTS_ETH_HDR_REG + ( resource_handle * PER_CONNECTION_ANTS_ETH_HDR_MEM_SIZE ) ;
  ipv4_hdr_mem_addr = SF_ANTS_IPV4_HDR_REG + ( resource_handle * PER_CONNECTION_ANTS_IPV4_HDR_MEM_SIZE ) ;
  tcp_hdr_mem_addr  = SF_ANTS_TCP_HDR_REG + ( resource_handle * PER_CONNECTION_ANTS_TCP_HDR_MEM_SIZE ) ;
  seq_num_mem_addr  = SF_SEQ_MAX_SEQ_NUM_REG + ( resource_handle * PER_CONNECTION_SEQ_AND_SEQ_NUM_MAX_SIZE );
  ants_tcp_rx_table_mem_addr = SF_ANTS_TCP_RX_TABLE_REG + ( resource_handle * PER_CONNECTION_ANTS_TCP_RX_TABLE_SIZE );


  /*Eth header DA/SA configuration*/
  status = ants_write_reg_block( handle, eth_hdr_mem_addr, ETH_DA_SA_SIZE, tcp_connection->header );
  if(status != AMM_SUCCESS) goto error;

  /*Ether_type configuration*/
  status = ants_write_reg(handle, eth_hdr_mem_addr + ETHER_TYPE_OFFSET, *((uint16_t*)(tcp_connection->header + ETH_DA_SA_SIZE)));
  if(status != AMM_SUCCESS) goto error;

  /*IPv4 header configuration*/
  status = ants_write_reg_block(handle, ipv4_hdr_mem_addr, (WORD_SIZE * 2), tcp_connection->header + (tcp_connection->ip_header_offset));
  if(status != AMM_SUCCESS) goto error;

  /*IPv4 TTL,PROTO,CSUM configuration*/
  data = *((uint16_t *)(tcp_connection->header + (tcp_connection->ip_header_offset + (WORD_SIZE * 2))));
  data |= (tcp_connection->ip_header_checksum << LEFT_SHIFT_LOW_TWO_BYTE);

  status = ants_write_reg(handle, ipv4_hdr_mem_addr + (WORD_SIZE * 2), data);
  if(status != AMM_SUCCESS) goto error;

  /*IPv4 DA/SA configuration*/
  status = ants_write_reg_block(handle, ipv4_hdr_mem_addr + (WORD_SIZE * 3) , (WORD_SIZE * 2), tcp_connection->header + (tcp_connection->ip_header_offset + (WORD_SIZE * 3)));
  if(status != AMM_SUCCESS) goto error;

  /*TCP src/dest port configuration*/
  status = ants_write_reg(handle, tcp_hdr_mem_addr, *((uint32_t*)(tcp_connection->header + (tcp_connection->tcp_header_offset))));
  if(status != AMM_SUCCESS) goto error;

  /*TCP ACK and SEQ configuration*/
  status = ants_write_reg(handle, tcp_hdr_mem_addr + WORD_SIZE , seq);
  if(status != AMM_SUCCESS) goto error;

  status = ants_write_reg(handle, tcp_hdr_mem_addr + (WORD_SIZE * 2), ack);
  if(status != AMM_SUCCESS) goto error;

  /*TCP window configuration*/
  data = 0;
  data = *((uint16_t *)(tcp_connection->header + (tcp_connection->tcp_header_offset + (WORD_SIZE * 3))));

  status = ants_write_reg(handle, tcp_hdr_mem_addr + (WORD_SIZE * 3) , data);
  if(status != AMM_SUCCESS) goto error;

  /*TCP csum configuration*/
  data = 0;
  data = *((uint16_t *)(tcp_connection->header + (tcp_connection->tcp_header_offset + (WORD_SIZE * 4) + 2)));

  status = ants_write_reg(handle, tcp_hdr_mem_addr + (WORD_SIZE * 4), data);
  if(status != AMM_SUCCESS) goto error;

  /*Seq_num configuration*/
  status = ants_write_reg(handle, seq_num_mem_addr + SEQ_NUM_OFFSET, tcp_connection->seq_num);
  if(status != AMM_SUCCESS) goto error;

  /*Seq_num_max configuration*/
  status = ants_write_reg(handle, seq_num_mem_addr, tcp_connection->seq_num_max);
  if(status != AMM_SUCCESS) goto error;

  /*ACK, window and tcp checksum configuration*/
  data=0;
  data = *((uint16_t*)(tcp_connection->header + tcp_connection->tcp_header_offset + TCP_WINDOW_SIZE_OFFSET));
  data |= (tcp_connection->tcp_header_checksum << LEFT_SHIFT_LOW_TWO_BYTE);

  status = ants_write_reg(handle, ants_tcp_rx_table_mem_addr , *((uint32_t*)(tcp_connection->header + tcp_connection->tcp_header_offset + TCP_ACK_OFFSET))) ;
  if(status != AMM_SUCCESS) goto error;

  status = ants_write_reg(handle, ants_tcp_rx_table_mem_addr + TCP_WINDOW_OFFSET , data);
  if(status != AMM_SUCCESS) goto error;

error:
  return status;

}

/**
 * @brief Enable ANTS header lookup
 *
 * @param handle <amm_handle_t> Handle to the device
 * @param ip_proto <uint8_t> IPv4 protocol
 * @param ip_address <uint32_t> IPv4 Destination address
 * @param port_num <uint16_t> UDP Destination port
 * @param enable <bool> Header match Enable/Disable
 *
 * @return AMM_SUCCESS or different error codes
 */
amm_err_t antshal_header_lu_enable (amm_handle_t handle, uint8_t ip_proto, uint32_t ip_dest_address, uint16_t port_num, bool enable)
{
  amm_err_t status = AMM_SUCCESS;

  /* Parameter Validation */
  if(NULL == handle)
    return AMM_ERROR_NOT_OPEN;

  if(!ip_proto || !ip_dest_address || !port_num)
    return AMM_ERROR_INVALID_PARAM;

  /* Header IPv4 Protocol configuration*/
  status = ants_write_reg(handle, SF_DS_RX_MEM_BASE + SF_HDR_IPV4_PROTO_OFFSET, ip_proto);
  if(status != AMM_SUCCESS) goto error;

  /* Header IPv4 Destination Address configuration*/
  status = ants_write_reg(handle, SF_DS_RX_MEM_BASE + SF_HDR_IPV4_DEST_ADDR_OFFSET, ip_dest_address);
  if(status != AMM_SUCCESS) goto error;

  /* Header L4 Destination Port configuration*/
  status = ants_write_reg(handle, SF_DS_RX_MEM_BASE + SF_HDR_L4_DEST_PORT_OFFSET, port_num);
  if(status != AMM_SUCCESS) goto error;

  /* Header Match Enable/Disable configuration*/
  status = ants_write_reg(handle, SF_DS_RX_MEM_BASE + SF_HDR_IPV4_ENABLE_OFFSET, enable);
  if(status != AMM_SUCCESS) goto error;

#if 1
  // initialize Impulse C module
#define MCAST_IP     0xe0010203
#define MCAST_PROTO  17
#define MCAST_PORT   8122

#define STREAM_CMD_MAGIC_NUMBER     0xABCDEF00
#define STREAM_CMD_PING             0
#define STREAM_CMD_WREG_IPV4DEST    1
#define STREAM_CMD_WREG_IPV4PROTO   2
#define STREAM_CMD_WREG_UDPPORT     3

  uint32_t data, dataRead;

#if 1
  data = STREAM_CMD_MAGIC_NUMBER | STREAM_CMD_PING;
  status = ants_write_reg(handle, 0x30000000, data);
  if(status != AMM_SUCCESS) goto error;
  data = 0x01234567;
  status = ants_write_reg(handle, 0x30000000, data);
  if(status != AMM_SUCCESS) goto error;

  status = ants_read_reg(handle, 0x30000000, &dataRead);
  if(status != AMM_SUCCESS) goto error;

  printf("antshal_header_lu_enable:Pinged hardware with 0x%08x, ping response=0x%08x\n",data,dataRead);
#endif

  data = STREAM_CMD_MAGIC_NUMBER | STREAM_CMD_WREG_IPV4DEST;
  status = ants_write_reg(handle, 0x30000000, data);
  if(status != AMM_SUCCESS) goto error;
  data = MCAST_IP;
  status = ants_write_reg(handle, 0x30000000, data);
  if(status != AMM_SUCCESS) goto error;
  data = STREAM_CMD_MAGIC_NUMBER | STREAM_CMD_WREG_IPV4PROTO;
  status = ants_write_reg(handle, 0x30000000, data);
  if(status != AMM_SUCCESS) goto error;
  data = MCAST_PROTO;
  status = ants_write_reg(handle, 0x30000000, data);
  if(status != AMM_SUCCESS) goto error;
  data = STREAM_CMD_MAGIC_NUMBER | STREAM_CMD_WREG_UDPPORT;
  status = ants_write_reg(handle, 0x30000000, data);
  if(status != AMM_SUCCESS) goto error;
  data = MCAST_PORT;
  status = ants_write_reg(handle, 0x30000000, data);
  if(status != AMM_SUCCESS) goto error;
#endif
error:
  return status;
}

/**
 * @brief Insert Prefix
 *
 * @param handle <amm_handle_t> Handle to the device
 * @param prefix <uint8_t> Prefix bytes that will be inserted after metadat in packet
 * @param prefix_len <uint32_t> Length of prefix
 *
 * @return AMM_SUCCESS or different error codes
 */
amm_err_t antshal_insert_prefix(amm_handle_t handle, uint8_t *prefix, uint32_t prefix_len)
{
  amm_err_t status = AMM_SUCCESS;

  /* Parameter Validation */
  if(NULL == handle)
    return AMM_ERROR_NOT_OPEN;

  if(NULL == prefix)
    return AMM_ERROR_INVALID_PARAM;

  if(prefix_len > FIXED_PREFIX_LENGTH)
    return AMM_ERROR_INVALID_PARAM;

  status = ants_write_reg_block(handle, SF_HDR_PREFIX_OFFSET, (prefix_len + SF_HDR_PREFIX_RESERVED_BYTES), prefix);

  return status;
}

/**
 * @brief ANTS message template configuration
 *
 * @param handle <amm_handle_t> Handle to the device
 * @param resource_handle <uint32_t> Message Id
 * @param msg_template <void*> Pointer to message template
 *
 * @return AMM_SUCCESS or different error codes
 */
amm_err_t antshal_message_template_config(amm_handle_t handle, uint32_t resource_handle, void *msg_temp)
{
  uint8_t *buffer = NULL;
  uint32_t msg_len_csum = 0;
  int32_t padding_bytes = 0;
  int32_t last_bytes_to_configure = 0;
  ants_message_template_info_t *msg_template = NULL;
  amm_err_t status = AMM_SUCCESS;

   /* Parameter Validation */
   if(NULL == handle)
    return AMM_ERROR_NOT_OPEN;

   if((NULL == msg_temp) || ((int32_t)resource_handle < 0))
    return AMM_ERROR_INVALID_PARAM;


  msg_template = (ants_message_template_info_t*) msg_temp;
  buffer = msg_template->msg_buffer;
  msg_len_csum = msg_template->msg_len;
  last_bytes_to_configure = ( msg_template->msg_len % SF_HW_BLOCK_REG_WIDTH);

   if(msg_template->msg_len > MSG_PAYLOAD_MAX_SIZE)
    return AMM_ERROR_INVALID_PARAM;

  /* calculation of padding bytes since the word size of 4 needs to
   * maintained while writing into FPGA memory
   */
  if(last_bytes_to_configure)
    padding_bytes = SF_HW_BLOCK_REG_WIDTH - last_bytes_to_configure;

  msg_len_csum = ((msg_len_csum << MSG_LEN_MASK) | msg_template->msg_checksum) ;

  status = ants_write_reg(handle, SF_MSG_LEN_CSUM_REG + (resource_handle * MSG_LEN_CSUM_MEM_SIZE), msg_len_csum);
   if(status != AMM_SUCCESS) goto error;

  status = ants_write_reg_block(handle,SF_MSG_BODY_REG + (resource_handle * MSG_BODY_MEM_SIZE), msg_template->msg_len + padding_bytes, buffer);
   if(status != AMM_SUCCESS) goto error;

  error:
    return status;
}

/**
 * @brief TCP connection Enable/Disable
 *
 * @param handle <amm_handle_t> Handle to the device
 * @param resource_handle <uint32_t> Connection Id
 * @param enable <bool> Enable/Disable TCP connection
 *
 * @return AMM_SUCCESS or different error codes
 */
amm_err_t antshal_tcp_connection_enable_disable(amm_handle_t handle, uint32_t resource_handle, bool enable)
{
  amm_err_t status = AMM_SUCCESS;

  /* Parameter Validation */
  if(NULL == handle)
    return AMM_ERROR_NOT_OPEN;

  if((int32_t)resource_handle < 0)
    return AMM_ERROR_INVALID_PARAM;

  /*TCP connection enable/disable configuration*/
  status = ants_write_reg(handle, SF_SEQ_MAX_SEQ_NUM_REG + (resource_handle * PER_CONNECTION_SEQ_AND_SEQ_NUM_MAX_SIZE) + TCP_ENABLE_DISABLE_OFFSET, enable);

  return status;
}
