--------------------------------------------------------------------------
-- Copyright (c) 2012 by Impulse Accelerated Technologies, Inc.    --
-- All rights reserved.                                                 --
--                                                                      --
-- This source file may be used and redistributed without charge        --
-- subject to the provisions of the IMPULSE ACCELERATED TECHNOLOGIES,   --
-- INC. REDISTRIBUTABLE IP LICENSE AGREEMENT, and provided that this    --
-- copyright statement is not removed from the file, and that any       --
-- derivative work contains this copyright notice.                      -
--------------------------------------------------------------------------
--                                                                      --
-- avalon_if.vhd: Altera Avalon stream, signal, and DMA interface       --
--   modules                                                            --
--                                                                      --
-- Change History
-- --------------
-- 9/20/2003 - Scott Thibault
--   File created.
-- 12/11/2003 - Scott Thibault
--   Added signal interface.
-- 2/20/2004 - Scott Thibault
--   Added Avalon DMA
-- 4/21/2004 - Scott Thibault
--   Added register interface.
--
--------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;

entity avalon_to_stream is
  generic (
    datawidth : positive := 8;
	 BUSDATAWIDTH : positive := 64 
  );
  port (
    clk : in std_logic;
    avalon_addr : in std_logic_vector (1 downto 0);
    avalon_ce : in std_logic;
    avalon_read_n : in std_logic;
    avalon_reset_n : in std_logic;
    avalon_write_n : in std_logic;
    avalon_wdata : in std_logic_vector ((BUSDATAWIDTH - 1) downto 0);
    avalon_rdata : out std_logic_vector ((BUSDATAWIDTH - 1) downto 0);
    stream_en : out std_ulogic;
    stream_rdy : in std_ulogic;
    stream_eos : out std_ulogic;
    stream_data : out std_ulogic_vector (datawidth-1 downto 0)
  );
end avalon_to_stream;

architecture rtl of avalon_to_stream is
  signal error, write_addr : std_ulogic;
  signal status : std_logic_vector ((BUSDATAWIDTH - 1) downto 0);
begin
  write_addr <=
    '1' when avalon_addr = "00" else
    '1' when avalon_addr = "10" else
    '0';

  -- Write to stream
  stream_en <= stream_rdy and avalon_ce and not avalon_write_n and write_addr;
  stream_eos <= '1' when avalon_addr = "10" else '0';
  stream_data <= std_ulogic_vector(avalon_wdata(datawidth-1 downto 0));

  -- Error detection
  check: process (avalon_reset_n, clk)
  begin
    if (avalon_reset_n = '0') then
      error <= '0';
    elsif (clk'event and clk='1') then
      if (avalon_ce = '1' and avalon_write_n = '0' and write_addr = '1') then
        error <= not stream_rdy;
      end if;
    end if;
  end process;
  
  -- Status register
  status(0) <= stream_rdy;
  status(1) <= '0';
  status(2) <= '0';
  status(3) <= error;
  status((BUSDATAWIDTH - 1) downto 4) <= X"000000000000000";

  avalon_rdata <=
    status when avalon_addr = "01" else
    X"0000000000000000";
end rtl;

library ieee;
use ieee.std_logic_1164.all;

entity stream_to_avalon is
  generic (
    datawidth : positive := 8;
	 BUSDATAWIDTH : positive := 64
  );
  port (
    clk : in std_logic;
    avalon_addr : in std_logic_vector (1 downto 0);
    avalon_ce : in std_logic;
    avalon_read_n : in std_logic;
    avalon_reset_n : in std_logic;
    avalon_write_n : in std_logic;
    avalon_wdata : in std_logic_vector ((BUSDATAWIDTH - 1) downto 0);
    avalon_rdata : out std_logic_vector ((BUSDATAWIDTH - 1) downto 0);
    stream_en : out std_ulogic;
    stream_rdy : in std_ulogic;
    stream_eos : in std_ulogic;
    stream_data : in std_ulogic_vector (datawidth-1 downto 0)
  );
end stream_to_avalon;

architecture rtl of stream_to_avalon is
  signal accept, valid, eos, error, address_stream : std_ulogic;
  signal status : std_logic_vector ((BUSDATAWIDTH - 1) downto 0);
  signal data : std_ulogic_vector ((BUSDATAWIDTH - 1) downto 0);
  signal extended_data : std_ulogic_vector ((BUSDATAWIDTH - 1) downto 0);
begin
  address_stream <= '1' when avalon_addr = "00" else '0';
  -- Read register
  accept <= (not valid or (avalon_ce and not avalon_read_n and address_stream));
  extended_data((BUSDATAWIDTH - 1) downto datawidth) <= (others => '0');
  extended_data(datawidth - 1 downto 0) <= stream_data;

  read_stream: process (avalon_reset_n, clk)
  begin
    if (avalon_reset_n = '0') then
      valid <= '0';
      eos <= '0';
    elsif (clk'event and clk='1') then
      if (stream_rdy = '1' and accept = '1') then
        data <= extended_data;
        eos <= stream_eos;
        valid <= '1';
      elsif (avalon_ce = '1' and avalon_read_n = '0' and address_stream = '1') then
        valid <= '0';
      end if;
    end if;
  end process;

  stream_en <= accept;

  -- Error detection
  check: process (avalon_reset_n, clk)
  begin
    if (avalon_reset_n = '0') then
      error <= '0';
    elsif (clk'event and clk='1') then
      if (avalon_ce = '1' and avalon_read_n = '0' and address_stream = '1') then
        error <= not valid;
      end if;
    end if;
  end process;
  
  -- Status register
  status(0) <= valid;
  status(1) <= eos;
  status(2) <= '0';
  status(3) <= error;
  status((BUSDATAWIDTH - 1) downto 4) <= X"000000000000000";

  avalon_rdata <=
    std_logic_vector(data) when avalon_addr = "00" else
    status when avalon_addr = "01" else
    X"0000000000000000";
end rtl;

library ieee;
use ieee.std_logic_1164.all;

entity avalon_to_signal is
  generic (
    datawidth : positive := 32;
	 BUSDATAWIDTH : positive := 64
  );
  port (
    clk : in std_logic;
    avalon_addr : in std_logic_vector (1 downto 0);
    avalon_ce : in std_logic;
    avalon_read_n : in std_logic;
    avalon_reset_n : in std_logic;
    avalon_write_n : in std_logic;
    avalon_wdata : in std_logic_vector ((BUSDATAWIDTH - 1) downto 0);
    avalon_rdata : out std_logic_vector ((BUSDATAWIDTH - 1) downto 0);
    signal_en : out std_ulogic;
    signal_data : out std_ulogic_vector (datawidth-1 downto 0)
  );
end avalon_to_signal;

architecture rtl of avalon_to_signal is
  signal write_addr : std_ulogic;
  signal status : std_logic_vector ((BUSDATAWIDTH - 1) downto 0);
begin
  write_addr <= '1' when avalon_addr = "00" else '0';

  -- Write to signal
  signal_en <= avalon_ce and not avalon_write_n and write_addr;
  signal_data <= std_ulogic_vector(avalon_wdata(datawidth-1 downto 0));
 
  -- Status register
  status <= X"0000000000000001"; -- Always ready/no errors.

  avalon_rdata <=
    status when avalon_addr = "01" else
    X"0000000000000000";
end rtl;

library ieee;
use ieee.std_logic_1164.all;

entity signal_to_avalon is
  generic (
    datawidth : positive := 32;
	 BUSDATAWIDTH : positive := 64
  );
  port (
    clk : in std_logic;
    avalon_addr : in std_logic_vector (1 downto 0);
    avalon_ce : in std_logic;
    avalon_read_n : in std_logic;
    avalon_reset_n : in std_logic;
    avalon_write_n : in std_logic;
    avalon_wdata : in std_logic_vector ((BUSDATAWIDTH - 1) downto 0);
    avalon_rdata : out std_logic_vector ((BUSDATAWIDTH - 1) downto 0);
    signal_en : out std_ulogic;
    signal_rdy : in std_ulogic;
    signal_data : in std_ulogic_vector (datawidth-1 downto 0)
  );
end signal_to_avalon;

architecture rtl of signal_to_avalon is
  signal accept, valid, error, address_signal : std_ulogic;
  signal status : std_logic_vector ((BUSDATAWIDTH - 1) downto 0);
  signal data : std_ulogic_vector ((BUSDATAWIDTH - 1) downto 0);
  signal extended_data : std_ulogic_vector ((BUSDATAWIDTH - 1) downto 0);
begin
  address_signal <= '1' when avalon_addr = "00" else '0';
  -- Read register
  accept <= '1';
  extended_data((BUSDATAWIDTH - 1) downto datawidth) <= (others => '0');
  extended_data(datawidth - 1 downto 0) <= signal_data;

  read_stream: process (avalon_reset_n, clk)
  begin
    if (avalon_reset_n = '0') then
      valid <= '0';
    elsif (clk'event and clk='1') then
      if (signal_rdy = '1' and accept = '1') then
        data <= extended_data;
        valid <= '1';
      elsif (avalon_ce = '1' and avalon_read_n = '0' and address_signal = '1') then
        valid <= '0';
      end if;
    end if;
  end process;

  signal_en <= accept;

  -- Error detection
  check: process (avalon_reset_n, clk)
  begin
    if (avalon_reset_n = '0') then
      error <= '0';
    elsif (clk'event and clk='1') then
      if (avalon_ce = '1' and avalon_read_n = '0' and address_signal = '1') then
        error <= not valid;
      end if;
    end if;
  end process;
  
  -- Status register
  status(0) <= valid;
  status(1) <= '0';
  status(2) <= '0';
  status(3) <= error;
  status((BUSDATAWIDTH - 1) downto 4) <= X"000000000000000";

  avalon_rdata <=
    std_logic_vector(data) when avalon_addr = "00" else
    status when avalon_addr = "01" else
    X"0000000000000000";
end rtl;

library ieee;
use ieee.std_logic_1164.all;

entity avalon_to_register is
  generic (
    datawidth : positive := 32;
	 BUSDATAWIDTH : positive := 64
  );
  port (
    clk : in std_logic;
    avalon_addr : in std_logic_vector (1 downto 0);
    avalon_ce : in std_logic;
    avalon_read_n : in std_logic;
    avalon_reset_n : in std_logic;
    avalon_write_n : in std_logic;
    avalon_wdata : in std_logic_vector ((BUSDATAWIDTH - 1) downto 0);
    avalon_rdata : out std_logic_vector ((BUSDATAWIDTH - 1) downto 0);
    register_en : out std_ulogic;
    register_data : out std_ulogic_vector (datawidth-1 downto 0)
  );
end avalon_to_register;

architecture rtl of avalon_to_register is
  signal write_addr : std_ulogic;
  signal status : std_logic_vector ((BUSDATAWIDTH - 1) downto 0);
begin
  write_addr <= '1' when avalon_addr = "00" else '0';

  -- Write to signal
  register_en <= avalon_ce and not avalon_write_n and write_addr;
  register_data <= std_ulogic_vector(avalon_wdata(datawidth-1 downto 0));
 
  -- Status register
  status <= X"0000000000000001"; -- Always ready/no errors.

  avalon_rdata <=
    status when avalon_addr = "01" else
    X"0000000000000000";
end rtl;

library ieee;
use ieee.std_logic_1164.all;

entity register_to_avalon is
  generic (
    datawidth : positive := 32;
	 BUSDATAWIDTH : positive := 64
  );
  port (
    clk : in std_logic;
    avalon_addr : in std_logic_vector (1 downto 0);
    avalon_ce : in std_logic;
    avalon_read_n : in std_logic;
    avalon_reset_n : in std_logic;
    avalon_write_n : in std_logic;
    avalon_wdata : in std_logic_vector ((BUSDATAWIDTH - 1) downto 0);
    avalon_rdata : out std_logic_vector ((BUSDATAWIDTH - 1) downto 0);
    register_value : in std_ulogic_vector (datawidth-1 downto 0)
  );
end register_to_avalon;

architecture rtl of register_to_avalon is
  signal extended_data : std_ulogic_vector ((BUSDATAWIDTH - 1) downto 0);
  signal status : std_logic_vector ((BUSDATAWIDTH - 1) downto 0);
begin
  extended_data((BUSDATAWIDTH - 1) downto datawidth) <= (others => '0');
  extended_data(datawidth - 1 downto 0) <= register_value;

  -- Status register
  status <= X"0000000000000001"; -- Always ready/no errors.

  avalon_rdata <=
    std_logic_vector(extended_data) when avalon_addr = "00" else
    status when avalon_addr = "01" else
    X"0000000000000000";
end rtl;

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

ENTITY avalon_dma IS
  PORT (
    reset, sclk, clk : in std_ulogic;
    AVaddr : out std_logic_vector (31 downto 0);
    AVben : out std_logic_vector (7 downto 0);
    AVread : out std_logic;
    AVreaddata : in std_logic_vector (63 downto 0);
    AVreaddata_valid : in std_logic;
    AVwrite : out std_logic;
    AVwritedata : out std_logic_vector (63 downto 0);
    AVwaitreq : in std_logic;
    ICidata : out std_ulogic_vector (63 downto 0);
    ICaddr : out std_ulogic_vector (31 downto 0);
    ICnextaddr : out std_ulogic_vector (31 downto 0);
    ICwri : out std_ulogic;
    ICre : out std_ulogic;              -- level 2
    ICodata : in std_ulogic_vector (63 downto 0);
    ICack : out std_ulogic;
    ICreq : in std_ulogic;
    ICblock : in std_ulogic;            -- level 2
    ICslice : in std_ulogic;            -- level 2
    ICmode : in std_ulogic;
    ICbase : in std_ulogic_vector (31 downto 0);
    ICsize : in std_ulogic_vector (3 downto 0);
    ICchunk : in std_ulogic_vector(31 downto 0);        -- level 2
    ICstride : in std_ulogic_vector(31 downto 0);       -- level 2
    ICstart : in std_ulogic_vector(31 downto 0);        -- level 2
    ICcount : in std_ulogic_vector (31 downto 0)
    );
end avalon_dma;

-- Implements a 1-D DMA with pipelined read support
ARCHITECTURE rtl OF avalon_dma IS

  TYPE state_type IS (idle, writing, reading, flushing, sequence);

  SIGNAL state_reg   : state_type;
  SIGNAL block_sz_reg     : unsigned(31 downto 0);
  SIGNAL cmd_cnt_reg       : unsigned(31 downto 0);
  SIGNAL cmd_done     : std_logic;
  SIGNAL xfer_cnt_reg     : unsigned(31 downto 0);
  SIGNAL transaction_done   : std_logic;
  SIGNAL xfer_done   : std_logic;
  SIGNAL sequence_done   : std_logic;
  SIGNAL total_cnt_reg   : unsigned(31 downto 0);
  SIGNAL av_addr_reg       : unsigned(31 downto 0);
  SIGNAL av_seq_reg     : unsigned(31 downto 0);
  SIGNAL ic_addr_reg       : unsigned(31 downto 0);
  SIGNAL ic_nextaddr_sig    : unsigned(31 downto 0);
  SIGNAL ic_nextxfer_sig    : unsigned(31 downto 0);
  SIGNAL ben         : std_logic_vector(7 downto 0);

begin

  --------------------------------------------------------------------------
  -- process: fsm
  -- PURPOSE: Master finite state machine for the DMA.
  --------------------------------------------------------------------------
  fsm : process (clk, reset)

  begin
    if (reset='1') then
      state_reg <= idle;
    elsif rising_edge(clk) then
      case state_reg IS
        when idle =>
          if(ICreq='1') then
            if(ICmode='1') then
              state_reg <= writing;
            else
              state_reg <= reading;
            end if;
          end if;
        when writing =>
          if (xfer_done='1') then
            if (sequence_done='1') then
              state_reg <= idle;
            else
              state_reg <= sequence;
            end if;
          end if;
        when reading =>
          if (cmd_done='1') then
            if (xfer_done='0') then
              state_reg <= flushing;
            else
              if (sequence_done='1') then
                state_reg <= idle;
              else
                state_reg <= sequence;
              end if;
            end if;
          end if;
        when flushing =>
          if (xfer_done='1') then
            if (sequence_done='1') then
              state_reg <= idle;
            else
              state_reg <= sequence;
            end if;
          end if;
        when sequence =>
          if(ICmode='1') then
            state_reg <= writing;
          else
            state_reg <= reading;
          end if;
      end case;
    end if;
  end process fsm;


  -------------------------------------------------------------------------
  -- process: cnt
  -- PURPOSE: Command transfer, data transfer, and source and destination
  --          address counters.
  -------------------------------------------------------------------------
  cnt : process (clk, reset)

  begin
    if (reset='1') then
      cmd_cnt_reg <= (others => '0');
      xfer_cnt_reg <= (others => '0');
      total_cnt_reg <= (others => '0');
      block_sz_reg <= (others => '0');
      ic_addr_reg <= (others => '0');
      av_addr_reg <= (others => '0');
      av_seq_reg <= (others => '0');
    elsif rising_edge(clk) then
      case state_reg IS
        when idle =>
          av_addr_reg <= unsigned(ICbase);
          av_seq_reg <= unsigned(ICbase);
          cmd_cnt_reg <= To_unsigned(1, cmd_cnt_reg'LENGTH);
          xfer_cnt_reg <= To_unsigned(1, xfer_cnt_reg'LENGTH);
          total_cnt_reg <= To_unsigned(1, total_cnt_reg'LENGTH);
          ic_addr_reg <= unsigned(ICstart);
          if (ICslice='0') then
            block_sz_reg <= unsigned(ICcount);
          else
            block_sz_reg <= unsigned(ICchunk);
          end if;
        when sequence =>
          av_addr_reg <= av_seq_reg + unsigned(ICstride);
          cmd_cnt_reg <= To_unsigned(1, cmd_cnt_reg'LENGTH);
          xfer_cnt_reg <= To_unsigned(1, xfer_cnt_reg'LENGTH);
          av_seq_reg <= av_seq_reg + unsigned(ICstride);
        when writing =>
          if (AVwaitreq='0') then
            cmd_cnt_reg <= cmd_cnt_reg + 1;
            av_addr_reg <= av_addr_reg + resize(unsigned(ICsize), av_addr_reg'LENGTH);
            xfer_cnt_reg <= xfer_cnt_reg + 1;
            total_cnt_reg <= total_cnt_reg + 1;
            ic_addr_reg <= ic_nextxfer_sig;
          end if;
        when reading =>
          if (AVwaitreq='0') then
            cmd_cnt_reg <= cmd_cnt_reg + 1;
            av_addr_reg <= av_addr_reg + resize(unsigned(ICsize), av_addr_reg'LENGTH);
          end if;
          if (AVreaddata_valid='1') then
            xfer_cnt_reg <= xfer_cnt_reg + 1;
            total_cnt_reg <= total_cnt_reg + 1;
            ic_addr_reg <= ic_nextxfer_sig;
          end if;
        when flushing =>
          if (AVreaddata_valid='1') then
            xfer_cnt_reg <= xfer_cnt_reg + 1;
            total_cnt_reg <= total_cnt_reg + 1;
            ic_addr_reg <= ic_nextxfer_sig;
          end if;
      end case;
    end if;
  end process cnt;  

  -- notE This needs to be a signal so that the FPGA can read the first
  -- value during the idle cycle preceeding a write and data is
  -- ready during the first cycle of the write.
  -- notE ICnextaddr is unused during memory transfer -> FPGA
  ic_nextaddr_sig <=
    unsigned(ICstart) when state_reg=idle else
    ic_nextxfer_sig when AVwaitreq='0' else
    ic_addr_reg;
  ic_nextxfer_sig <= ic_addr_reg + resize(unsigned(ICsize),ic_addr_reg'LENGTH);
    
  cmd_done <=
    not AVwaitreq when cmd_cnt_reg=block_sz_reg AND state_reg/=idle else
    '0';
  transaction_done <=
    not AVwaitreq when state_reg=writing else
    AVreaddata_valid when state_reg=reading OR state_reg=flushing else
    '0';
  xfer_done <= transaction_done when xfer_cnt_reg=block_sz_reg else '0';
  sequence_done <=
    transaction_done when total_cnt_reg=unsigned(ICcount) else '0';
  
  -------------------------------------------------------------------------
  -- process: gen_ben
  -- PURPOSE: Generate byte enables on the avalon bus.
  -------------------------------------------------------------------------
  gen_ben : process(ICsize, av_addr_reg)

    VARIABLE size_int : integer RANGE 0 TO 15;
    VARIABLE addr_lsbs_int : integer RANGE 0 TO 7;

  begin
    size_int := To_integer(unsigned(ICsize));
    addr_lsbs_int := To_integer(unsigned(av_addr_reg(2 downto 0)));
    case size_int IS
      when 1 =>
        case addr_lsbs_int IS
          when 0 => ben <= "00000001";
          when 1 => ben <= "00000010";
          when 2 => ben <= "00000100";
          when 3 => ben <= "00001000";
          when 4 => ben <= "00010000";
          when 5 => ben <= "00100000";
          when 6 => ben <= "01000000";
          when 7 => ben <= "10000000";
        end case;
      when 2 =>
        case addr_lsbs_int IS
          when 0 | 1 => ben <= "00000011";
          when 2 | 3 => ben <= "00001100";
          when 4 | 5 => ben <= "00110000";
          when 6 | 7 => ben <= "11000000";
        end case;
      when 4 =>
        case addr_lsbs_int IS
          when 0 TO 3 => ben <= "00001111";
          when 4 TO 7 => ben <= "11110000";
        end case;                
      when others =>
        ben <= "11111111";
    end case;   
  end process gen_ben;

  -- ye olde avalon bus
  AVaddr <= std_logic_vector(av_addr_reg);
  AVben <= ben; -- always drive byte enables -- when state_reg=writing else "11111111";
  AVread  <= '1' when state_reg=reading else '0';
  AVwrite <= '1' when state_reg=writing else '0';
  AVwritedata <= std_logic_vector(ICodata);

  -- impulse c bus
  ICidata <= std_ulogic_vector(AVreaddata);
  ICwri <= AVreaddata_valid when state_reg=reading OR state_reg=flushing else '0';
  ICack <= sequence_done when state_reg/=idle else '0';
  ICaddr <= std_ulogic_vector(ic_addr_reg);
  ICnextaddr <= std_ulogic_vector(ic_nextaddr_sig);
  ICre <= not AVwaitreq when state_reg=writing else '0';
end rtl;

--------------------------------------------------------------------------
--   These adapters has been developed for Avalon Streams:
--   If not already done make a Megacore interface with the following 
--   Avalon ST options:
--      readyLatency = 0
--      dataBitsPerSymbol = 8 
--      symbolsPerBeat    = (set the datawidth/8)
--      maxChannels       = 0 (may need to use a channel adapter)
--------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;

entity avalon_st_pkt_to_stream is
	generic (
		datawidth : positive := 32
	);
	port (
		clk : in std_logic;
		reset : in std_logic;
		avalon_ready : out std_logic;
		avalon_valid : in std_logic;
		avalon_endofpacket : in std_logic;
		avalon_startofpacket : in std_logic;
		avalon_data : in std_logic_vector (datawidth-1 downto 0);
		stream_en : out std_ulogic;
		stream_rdy : in std_ulogic;
		stream_eos : out std_ulogic;
		stream_data : out std_ulogic_vector (datawidth-1 downto 0)
	);
end avalon_st_pkt_to_stream;

architecture rtl of avalon_st_pkt_to_stream is
	signal endofstream : std_ulogic;
begin
	
	stream_data <= std_ulogic_vector(avalon_data);
	stream_en <= avalon_valid or endofstream;
	avalon_ready <= stream_rdy and (not endofstream);
  
	process (clk,reset,avalon_endofpacket)
	begin
		if (reset='1') then
			endofstream <= '0';
		elsif (clk'event and clk='1') then
			if (stream_rdy = '1' ) then
				if (avalon_endofpacket = '1' and avalon_valid = '1') then
					endofstream <= '1';
				else
					endofstream <= '0';
				end if;
			end if;
		end if;
	end process;
	
	stream_eos <= endofstream;

end rtl;

library ieee;
use ieee.std_logic_1164.all;

entity stream_to_avalon_st_pkt is
	generic (
		datawidth : positive := 32
	);
	port (
		clk : in std_logic;
		reset : in std_logic;
		avalon_valid : out std_logic;
		avalon_ready : in std_logic;
		avalon_endofpacket : out std_logic;
		avalon_startofpacket : out std_logic;
		avalon_data : out std_logic_vector (datawidth-1 downto 0);
		stream_en : out std_ulogic;
		stream_rdy : in std_ulogic;
		stream_eos : in std_ulogic;
		stream_data : in std_ulogic_vector (datawidth-1 downto 0)
	);
end stream_to_avalon_st_pkt;

architecture rtl of stream_to_avalon_st_pkt is
	signal packet_open : std_ulogic;
	signal start_of_packet : std_ulogic;
	signal data : std_ulogic_vector (datawidth-1 downto 0);
	signal data_valid : std_ulogic;
	signal do_buf_xfer : std_ulogic;
	signal do_av_xfer : std_ulogic;
begin

	-- need to look ahead one word in order to combine EOS with last word
	--  so add one word buffer/register - it now takes two words preset or one word
	--  and EOS present to do a transfer
	do_av_xfer <= avalon_ready;
	do_buf_xfer <= ((data_valid and do_av_xfer) or (not data_valid)) and stream_rdy; 
	stream_en <= do_buf_xfer; 
	read_buff: process(clk, reset)
	begin
		if (reset = '1') then
			data_valid <= '0';
			packet_open <= '0';
		elsif (rising_edge(clk)) then
			if (do_buf_xfer = '1') then
				data <= stream_data;
				data_valid <= not stream_eos;
				if (data_valid = '0' and packet_open = '0') then
					packet_open <= '1';
					start_of_packet <= '1';
				elsif (stream_eos  = '1' ) then
					packet_open <= '0';
				else
					start_of_packet <= '0';
				end if;
			end if;
		end if;
	end process;
	
	avalon_valid <= do_buf_xfer and data_valid;
	avalon_startofpacket <= start_of_packet;
	avalon_endofpacket <= stream_eos;
	avalon_data <= std_logic_vector(data);
	
end rtl;

-- Avalon ST "avalon_st" <-> co_no_eos_stream:
--		- Intended for raw access to protocol signals via co_stream data 
-- 	- does not support EOS relying on EOP and other signals to be detected 
--			as part of the data stream
--		- pass packet signals SOP, EOP, and others Empty signals as part of the data stream:
--			- this is enabled in SOPC Builder as an option to include SOP+EOP in 
--				stream data and the generated instance/adapter actually does the signal 
--				multiplexing and demultiplexing

library ieee;
use ieee.std_logic_1164.all;

entity avalon_st_to_no_eos_stream is
	generic (
		datawidth : positive := 32
	);
	port (
		clk : in std_logic;
		reset : in std_logic;
		avalon_ready : out std_logic;
		avalon_valid : in std_logic;
		avalon_data : in std_logic_vector (datawidth-1 downto 0);
		stream_en : out std_ulogic;
		stream_rdy : in std_ulogic;
		stream_eos : out std_ulogic;
		stream_data : out std_ulogic_vector (datawidth-1 downto 0)
	);
end avalon_st_to_no_eos_stream;

architecture rtl of avalon_st_to_no_eos_stream is
begin
	stream_data <= std_ulogic_vector(avalon_data);
	stream_en <= avalon_valid;
	avalon_ready <= stream_rdy;
	stream_eos <= '0';
end rtl;

library ieee;
use ieee.std_logic_1164.all;

entity no_eos_stream_to_avalon_st is
	generic (
		datawidth : positive := 32
	);
	port (
		clk : in std_logic;
		reset : in std_logic;
		avalon_valid : out std_logic;
		avalon_ready : in std_logic;
		avalon_data : out std_logic_vector (datawidth-1 downto 0);
		stream_en : out std_ulogic;
		stream_rdy : in std_ulogic;
		stream_eos : in std_ulogic;
		stream_data : in std_ulogic_vector (datawidth-1 downto 0)
	);
end no_eos_stream_to_avalon_st;

architecture rtl of no_eos_stream_to_avalon_st is
begin
	stream_en <= avalon_ready;
	avalon_valid <= stream_rdy;
	avalon_data <= std_logic_vector(stream_data(datawidth-1 downto 0));
end rtl;

library ieee;
use ieee.std_logic_1164.all;

package avalon_if is
  component avalon_to_stream is
    generic (
      datawidth : positive := 8;
	   BUSDATAWIDTH : positive := 64
    );
    port (
      clk : in std_logic;
      avalon_addr : in std_logic_vector (1 downto 0);
      avalon_ce : in std_logic;
      avalon_read_n : in std_logic;
      avalon_reset_n : in std_logic;
      avalon_write_n : in std_logic;
      avalon_wdata : in std_logic_vector ((BUSDATAWIDTH - 1) downto 0);
      avalon_rdata : out std_logic_vector ((BUSDATAWIDTH - 1) downto 0);
      stream_en : out std_ulogic;
      stream_rdy : in std_ulogic;
      stream_eos : out std_ulogic;
      stream_data : out std_ulogic_vector (datawidth-1 downto 0)
    );
  end component;

  component stream_to_avalon is
    generic (
      datawidth : positive := 8;
	   BUSDATAWIDTH : positive := 64
    );
    port (
      clk : in std_logic;
      avalon_addr : in std_logic_vector (1 downto 0);
      avalon_ce : in std_logic;
      avalon_read_n : in std_logic;
      avalon_reset_n : in std_logic;
      avalon_write_n : in std_logic;
      avalon_wdata : in std_logic_vector ((BUSDATAWIDTH - 1) downto 0);
      avalon_rdata : out std_logic_vector ((BUSDATAWIDTH - 1) downto 0);
      stream_en : out std_ulogic;
      stream_rdy : in std_ulogic;
      stream_eos : in std_ulogic;
      stream_data : in std_ulogic_vector (datawidth-1 downto 0)
    );
  end component;

  component avalon_to_signal is
    generic (
      datawidth : positive := 32;
	   BUSDATAWIDTH : positive := 64
    );
    port (
      clk : in std_logic;
      avalon_addr : in std_logic_vector (1 downto 0);
      avalon_ce : in std_logic;
      avalon_read_n : in std_logic;
      avalon_reset_n : in std_logic;
      avalon_write_n : in std_logic;
      avalon_wdata : in std_logic_vector ((BUSDATAWIDTH - 1) downto 0);
      avalon_rdata : out std_logic_vector ((BUSDATAWIDTH - 1) downto 0);
      signal_en : out std_ulogic;
      signal_data : out std_ulogic_vector (datawidth-1 downto 0)
    );
  end component;

  component signal_to_avalon is
    generic (
      datawidth : positive := 32;
	   BUSDATAWIDTH : positive := 64
    );
    port (
      clk : in std_logic;
      avalon_addr : in std_logic_vector (1 downto 0);
      avalon_ce : in std_logic;
      avalon_read_n : in std_logic;
      avalon_reset_n : in std_logic;
      avalon_write_n : in std_logic;
      avalon_wdata : in std_logic_vector ((BUSDATAWIDTH - 1) downto 0);
      avalon_rdata : out std_logic_vector ((BUSDATAWIDTH - 1) downto 0);
      signal_en : out std_ulogic;
      signal_rdy : in std_ulogic;
      signal_data : in std_ulogic_vector (datawidth-1 downto 0)
    );
  end component;

  component avalon_to_register is
    generic (
      datawidth : positive := 32;
	   BUSDATAWIDTH : positive := 64
    );
    port (
      clk : in std_logic;
      avalon_addr : in std_logic_vector (1 downto 0);
      avalon_ce : in std_logic;
      avalon_read_n : in std_logic;
      avalon_reset_n : in std_logic;
      avalon_write_n : in std_logic;
      avalon_wdata : in std_logic_vector ((BUSDATAWIDTH - 1) downto 0);
      avalon_rdata : out std_logic_vector ((BUSDATAWIDTH - 1) downto 0);
      register_en : out std_ulogic;
      register_data : out std_ulogic_vector (datawidth-1 downto 0)
    );
  end component;

  component register_to_avalon is
    generic (
      datawidth : positive := 32;
	   BUSDATAWIDTH : positive := 64
    );
    port (
      clk : in std_logic;
      avalon_addr : in std_logic_vector (1 downto 0);
      avalon_ce : in std_logic;
      avalon_read_n : in std_logic;
      avalon_reset_n : in std_logic;
      avalon_write_n : in std_logic;
      avalon_wdata : in std_logic_vector ((BUSDATAWIDTH - 1) downto 0);
      avalon_rdata : out std_logic_vector ((BUSDATAWIDTH - 1) downto 0);
      register_value : in std_ulogic_vector (datawidth-1 downto 0)
    );
  end component;

	component avalon_dma is
	  port (
		 reset, sclk, clk : in std_ulogic;
		 AVaddr : out std_logic_vector (31 downto 0);
		 AVben : out std_logic_vector (7 downto 0);
		 AVread : out std_logic;
		 AVreaddata : in std_logic_vector (63 downto 0);
		 AVreaddata_valid : in std_logic;
		 AVwrite : out std_logic;
		 AVwritedata : out std_logic_vector (63 downto 0);
		 AVwaitreq : in std_logic;
		 ICidata : out std_ulogic_vector (63 downto 0);
		 ICaddr : out std_ulogic_vector (31 downto 0);
		 ICnextaddr : out std_ulogic_vector (31 downto 0);
		 ICwri : out std_ulogic;
		 ICre : out std_ulogic;              -- level 2
		 ICodata : in std_ulogic_vector (63 downto 0);
		 ICack : out std_ulogic;
		 ICreq : in std_ulogic;
		 ICblock : in std_ulogic;            -- level 2
		 ICslice : in std_ulogic;            -- level 2
		 ICmode : in std_ulogic;
		 ICbase : in std_ulogic_vector (31 downto 0);
		 ICsize : in std_ulogic_vector (3 downto 0);
		 ICchunk : in std_ulogic_vector(31 downto 0);        -- level 2
		 ICstride : in std_ulogic_vector(31 downto 0);       -- level 2
		 ICstart : in std_ulogic_vector(31 downto 0);        -- level 2
		 ICcount : in std_ulogic_vector (31 downto 0)
		 );
	end component;
	
	component avalon_st_pkt_to_stream is
		generic (
			datawidth : positive := 32
		);
		port (
			clk : in std_logic;
			reset : in std_logic;
			avalon_ready : out std_logic;
			avalon_valid : in std_logic;
			avalon_endofpacket : in std_logic;
			avalon_startofpacket : in std_logic;
			avalon_data : in std_logic_vector (datawidth-1 downto 0);
			stream_en : out std_ulogic;
			stream_rdy : in std_ulogic;
			stream_eos : out std_ulogic;
			stream_data : out std_ulogic_vector (datawidth-1 downto 0)
		);
	end component;
	
	component stream_to_avalon_st_pkt is
		generic (
			datawidth : positive := 32
		);
		port (
			clk : in std_logic;
			reset : in std_logic;
			avalon_valid : out std_logic;
			avalon_ready : in std_logic;
			avalon_endofpacket : out std_logic;
			avalon_startofpacket : out std_logic;
			avalon_data : out std_logic_vector (datawidth-1 downto 0);
			stream_en : out std_ulogic;
			stream_rdy : in std_ulogic;
			stream_eos : in std_ulogic;
			stream_data : in std_ulogic_vector (datawidth-1 downto 0)
		);
	end component;
	
	component avalon_st_to_no_eos_stream is
		generic (
			datawidth : positive := 32
		);
		port (
			clk : in std_logic;
			reset : in std_logic;
			avalon_ready : out std_logic;
			avalon_valid : in std_logic;
			avalon_data : in std_logic_vector (datawidth-1 downto 0);
			stream_en : out std_ulogic;
			stream_rdy : in std_ulogic;
			stream_eos : out std_ulogic;
			stream_data : out std_ulogic_vector (datawidth-1 downto 0)
		);
	end component;
	
	component no_eos_stream_to_avalon_st is
		generic (
			datawidth : positive := 32
		);
		port (
			clk : in std_logic;
			reset : in std_logic;
			avalon_valid : out std_logic;
			avalon_ready : in std_logic;
			avalon_data : out std_logic_vector (datawidth-1 downto 0);
			stream_en : out std_ulogic;
			stream_rdy : in std_ulogic;
			stream_eos : in std_ulogic;
			stream_data : in std_ulogic_vector (datawidth-1 downto 0)
		);
	end component;

end avalon_if;

